/*
 * SPDX-FileCopyrightText: (C) 2022 Red Hat (www.redhat.com)
 * SPDX-License-Identifier: LGPL-2.1-or-later
 */

#ifndef E_RSS_PREFERENCES_H
#define E_RSS_PREFERENCES_H

#include <glib.h>

#include <shell/e-shell.h>

G_BEGIN_DECLS

void		e_rss_preferences_init	(EShell *shell);

G_END_DECLS

#endif /* E_RSS_PREFERENCES_H */
