/*
 * e-mail.h
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 * or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
 * for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 *
 */

#ifndef E_MAIL_H
#define E_MAIL_H

#include <mail/e-mail-backend.h>
#include <mail/e-mail-browser.h>
#include <mail/e-mail-display.h>
#include <mail/e-mail-folder-pane.h>
#include <mail/e-mail-label-dialog.h>
#include <mail/e-mail-label-list-store.h>
#include <mail/e-mail-label-manager.h>
#include <mail/e-mail-label-tree-view.h>
#include <mail/e-mail-message-pane.h>
#include <mail/e-mail-migrate.h>
#include <mail/e-mail-notes.h>
#include <mail/e-mail-paned-view.h>
#include <mail/e-mail-reader.h>
#include <mail/e-mail-reader-utils.h>
#include <mail/e-mail-sidebar.h>
#include <mail/e-mail-tag-editor.h>
#include <mail/e-mail-view.h>

#endif /* E_MAIL_H */
