((c-mode .  ((indent-tabs-mode . t)
             (c-basic-offset . 8)
             (tab-width . 8)))
 (js-mode . ((indent-tabs-mode . t)
             (c-basic-offset . 8)
             (tab-width . 8)))
 (auto-mode-alist . (("\\.eui\\'" . xml-mode)))
)
