:root {
  color-scheme: light dark; /* both supported */
}

html, body {
 padding: 0;
 margin: 0;
}

img {
  border: 0;
}

body {
  /* Use padding so that children can safely use width=100% */
  padding: 8px;
}

body, div, p, td {
  unicode-bidi: plaintext;
}

th {
  text-align: left;
  vertical-align: top;
}

span.navigable, div.navigable, p.navigable {
  cursor: pointer;
  text-decoration: underline;
}

img#__evo-contact-photo {
  max-width: 96px;
  max-height: 96px;
  border-radius: 50%;
}

img.navigable {
  cursor: pointer;
  padding-right: 4px;
}

.attachments {
  background: #FFF;
  border: 1px solid silver;
  padding: 10px 10px 10px 10px;
  border-left: 0;
  border-right: 0;
  border-bottom: 0;
}

.attachment {
  padding-left: 8px;
  padding-right: 0px;
}

.attachment td {
  padding: 0px;
}

iframe:not([id$=".itip"]) {
  border-width: 1px;
  border-style: solid;
}

.part-container {
  height: 100%;
  padding: 0px;
}

.part-container-nostyle iframe {
  margin: 0px;
  padding-right: 0px;
}

.part-container-inner-margin {
  padding: 0px;
}

object { /* GtkWidgets */
  padding-top: 2px;
  padding-bottom: 2px;
}

.__evo-highlight {
  color: purple;
  font-weight: bold;
}

a.evo-awrap {
  white-space: normal;
  word-break: break-all;
}

.calpreview th {
  white-space: nowrap;
  width: 1%;
}

.calpreview td {
  width: 99%;
}

/**** HEADERS ****/

button.header-collapse {
  width: 18px;
  height: 18px;
  margin: 0;
  padding: 0;
  cursor: pointer;
  overflow: hidden;
  display: flex;
  align-items: center;
  justify-content: center;
}

button.header-collapse img {
  width: 16px;
  height: 16px;
  padding: 0;
  margin: 0;
}

table.header {
  border-spacing: 0px;
  border-top-width: 0px;
  border-right-width: 0px;
  border-bottom-width: 0px;
  border-left-width: 0px;
  width: 100%;
}

tr.header {
  vertical-align: top;
  width: 100%;
}

th.header {
  white-space: nowrap;
}

td.header {
  width: inherit;
  white-space: normal;
  word-break: normal;
}

/* We don't want to break in the middle of contact. */
td.header a {
  white-space: nowrap;
}

.ltr {
  text-align: left;
}

th.ltr {
  padding-right: 6px;
  text-align: right;
  width: auto;
}

.rtl {
  text-align: right;
}

th.rtl {
  padding-left: 6px;
  text-align: left;
  width: auto;
}

/***** PRINTING *******/

.printing-header {
  padding-bottom: 20px;
}

.printing-header h1,
.attachments-list h1 {
  font-size: 20px;
}

.printing-header th {
  text-align: right;
  font-weight: bold;
}

.attachments-list th {
  font-weight: bold;
}

/******* ITIP *********/
.itip.icon {
  float: left;
  padding-right: 5px;
}

.itip.content {
  float: left;
  max-width: 90%;
}

.itip.description {
  padding: 5px;
}

.itip tr {
  vertical-align: middle;
  padding-top: 5px;
  padding-bottom: 5px;
}

.itip th {
  vertical-align: middle;
}

#table_row_summary td {
  font-weight: bold;
}

#table_row_buttons button, .table_row_buttons button {
  line-height: 28px;
  min-width: 150px;
  white-space: nowrap;
}

#table_row_buttons img, .table_row_buttons img {
  padding-right: 5px;
  vertical-align: middle;
}

#text_row_buttons td {
  text-align: center;
}
